/*
 *	xcardmon --- PCMCIA card utility for X11
 * Copyright (C) 1999 by ARAI Yoshishige <ryo2@jp.freebsd.org>
 * All right reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
#ifndef lint
static const char rcsid[] = "$Id: xcardmon.c,v 1.5 1999/06/19 17:55:43 ryo2 Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <pccard/cardinfo.h>
#include <pccard/cis.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/un.h>
#include <regex.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/cursorfont.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Cardinals.h>


#define APP_CLASS "XCardmon"
#define VERSION "0.5"

#define MY_FRAME_WIDTH		250

#define BG_COLOR		"lightcyan"
#define EMPTY_COLOR		"gray40"
#define ACTIVE_COLOR		"PaleGoldenrod"
#define HIBERNATE_COLOR		"gold3"
#define SWITCH_COLOR		"MediumSpringGreen"
#define POWER_COLOR		"orange"
#define SAFE_COLOR		"cyan3"

#define CONFIG_FILE		"/usr/local/lib/xpccard/config"
#define SERVER_NAME		"/var/tmp/.pccardd"
#define CLIENT_NAME		"/tmp/.xpccard"

#define CMD_NR_SLOTS		"S\n"
#define CMD_GET_INFO		"N%d\n"
#define CMD_PWR_ON		"P%d\n"
#define CMD_PWR_OFF		"Q%d\n"

#define MAX_PC_SLOT		4
#define MAX_INFO_LEN		64
#define MAX_CARD_NAME		MAX_INFO_LEN * 2
#define MAX_NAME_FEILD		5
#ifdef XAW3D
#define HORIZ_SPACE		10
#define LABEL_BORDER_WIDTH	2
#else
#define HORIZ_SPACE		12
#define LABEL_BORDER_WIDTH	1
#endif

struct SlotFrame {
    Widget box;
    Widget pwrbutton;
    Widget drvbutton;
    Widget drvinfo;
    Widget label;
};
static struct SlotFrame slotFrame[MAX_PC_SLOT];
static Display *disp;
static int screen, root_height, root_width;
static XtAppContext app_con;
#if XtSpecificationRelease >= 6
static XtSignalId sigHUPid, sigTERMid;
#endif
static Widget top, frame, bsb;
static Pixel statColor[4], PowerColor, SafeColor;
#ifdef I18N
#define XtNfontDEF XtNfontSet
XFontSet textfont, buttonfont;
#else
#define XtNfontDEF XtNfont
XFontStruct *textfont, *buttonfont;
#endif

static int SOCKET;
static int nr_slot = 0;
static int slen, alias_n = -1;
static char *myname;
static struct sockaddr_un saddr, caddr;
static char *S_NA = "N/A";
static char *S_HIBERNATING = "Hibernating Card";
static char *S_EMPTY = "    Empty    ";

struct CardInfo {
    int  stat;
    char manuf[MAX_INFO_LEN+4];
    char vers[MAX_INFO_LEN+4];
    char drv[MAX_INFO_LEN+4];
    char name[MAX_CARD_NAME+4];
};
static struct CardInfo cardInfo[MAX_PC_SLOT];

static struct io_desc cardIO[MAX_PC_SLOT];

struct CardAlias {
    char *alias;
    char *manuf;
    char *vers;
    char *buff;
    struct CardAlias *next;
};
static struct CardAlias *aliasTop = NULL;
static struct CardAlias *aliasPrev;

static void usage(FILE *, int);
static int  openSocket(void);
static void closeSocket(void);
static int  communicate(char *, char *, int);
static int  ismatch(char *, char*);
static char *nameAlias(struct CardAlias *, struct CardInfo *);
static int  parseCardName(char *, struct CardInfo *);
static int  getCardIOinfo(int, struct io_desc *);
static void getCardInfo(int, Bool);
static void setCardLabel(int, Bool);
static int  shellWord(char **, char **);
static int  readConfig(char *);
static void freeAliasArray(struct CardAlias **);
#if XtSpecificationRelease >= 6
static void doSigHUP(XtPointer, XtSignalId *);
static void doSigTERM(XtPointer, XtSignalId *);
#endif
static void signalHandler(int);

static void setAppData(void);
static void quitAction(Widget,XEvent *,String *,Cardinal *);
static void toggleSlot(Widget,XEvent *,String *,Cardinal *);
static void placeMenu(Widget,XButtonEvent *);
static void powerButton(Widget, XtPointer, XtPointer);
static void statChanged(Widget, int *, XtInputId *);
static void makeSlotFrame(int);
static void dumpname(int);
static void dumpalias(struct CardAlias *);

static struct _app_resources {
    Boolean drive_state, always_show, small, help, version, debug;
    String config_file;
    Pixel empty_color,active_color,hibernate_color;
    Pixel switch_color,power_color,safe_color;
#ifdef I18N
  XFontSet text_font, button_font;
#else
  XFontStruct *text_font, *button_font;
#endif
} opts;

static XtResource resources[] ={
#define offset(entry) XtOffset(struct _app_resources*, entry)
    {"sensitive", "Sensitive", XtRBoolean, sizeof(Boolean),
     offset(drive_state), XtRImmediate, (XtPointer)False},
    {"small", "Small", XtRBoolean, sizeof(Boolean),
     offset(small), XtRImmediate, (XtPointer)False},
    {"alwaysShow", "AlwaysShow", XtRBoolean, sizeof(Boolean),
     offset(always_show), XtRImmediate, (XtPointer)False},
    {"configFile", "ConfigFile", XtRString, sizeof(String),
     offset(config_file), XtRString, CONFIG_FILE},
    {"help", "Help", XtRBoolean, sizeof(Boolean),
     offset(help), XtRImmediate, (XtPointer)False},
    {"version", "Version", XtRBoolean, sizeof(Boolean),
     offset(version), XtRImmediate, (XtPointer)False},
    {"debug", "Debug", XtRBoolean, sizeof(Boolean),
     offset(debug), XtRImmediate, (XtPointer)False},
    {"emptyColor", "EmptyColor", XtRPixel, sizeof(Pixel),
     offset(empty_color), XtRString, EMPTY_COLOR},
    {"activeColor", "ActiveColor", XtRPixel, sizeof(Pixel),
     offset(active_color), XtRString, ACTIVE_COLOR},
    {"hibernateColor", "HibernateColor", XtRPixel, sizeof(Pixel),
     offset(hibernate_color), XtRString, HIBERNATE_COLOR},
    {"switchColor", "SwitchColor", XtRPixel, sizeof(Pixel),
     offset(switch_color), XtRString, SWITCH_COLOR},
    {"powerColor", "PowerColor", XtRPixel, sizeof(Pixel),
     offset(power_color), XtRString, POWER_COLOR},
    {"safeColor", "SafeColor", XtRPixel, sizeof(Pixel),
     offset(safe_color), XtRString, SAFE_COLOR},
#ifdef I18N
    {"fontSet", "FontSet", XtRFontSet, sizeof(XFontSet),
    offset(text_font), XtRString, "-*-fixed-medium-r-normal--14-*-*-*-*-*-*-*"},
    {"buttonFontSet", "ButtonFontSet", XtRFontSet, sizeof(XFontSet),
#ifdef XAW3D
   offset(button_font), XtRString, "-*-fixed-medium-r-normal--14-*-*-*-*-*-*-*"},
#else
   offset(button_font), XtRString, "-*-fixed-medium-r-normal--14-*-*-*-*-*-*-*"},
#endif
#else	/* !I18N */
    {"font", XtCFont, XtRFontStruct, sizeof(XFontStruct *),
     offset(text_font), XtRString, "-*-helvetica-medium-r-normal--12-*-*-*-*-*-*-*"},
    {"buttonFont", "ButtonFont", XtRFontStruct, sizeof(XFontStruct *),
#ifdef XAW3D
     offset(button_font), XtRString, "-*-helvetica-medium-r-normal--10-*-*-*-*-*-*-*"},
#else
     offset(button_font), XtRString, "-*-helvetica-medium-r-normal--12-*-*-*-*-*-*-*"},
#endif
#endif
#undef offset
};

XrmOptionDescRec options[] = {
    {"-sensitive", ".sensitive", XrmoptionNoArg, "True"},
    {"-s", ".sensitive", XrmoptionNoArg, "True"},
    {"-n", ".small", XrmoptionNoArg, "True"},
    {"-f", ".configFile", XrmoptionSepArg, NULL},
    {"-always", ".alwaysShow", XrmoptionNoArg, "True"},
#ifdef I18N
    {"-fs", ".fontSet", XrmoptionSepArg, NULL},
    {"-bfs", ".buttonFontSet", XrmoptionSepArg, NULL},
#else
    {"-bfn", ".buttonFont", XrmoptionSepArg, NULL},
#endif
    {"-help", ".help", XrmoptionNoArg, "True"},
    {"-h", ".help", XrmoptionNoArg, "True"},
    {"-v", ".version", XrmoptionNoArg, "True"},
    {"-d", ".debug", XrmoptionNoArg, "True"},
    {"-emptycolor", ".emptyColor", XrmoptionSepArg, NULL},
    {"-activecolor", ".activeColor", XrmoptionSepArg, NULL},
    {"-hibernatecolor", ".hibernateColor", XrmoptionSepArg, NULL},
    {"-switchcolor", ".switchColor", XrmoptionSepArg, NULL},
    {"-powercolor", ".powerColor", XrmoptionSepArg, NULL},
    {"-safecolor", ".safeColor", XrmoptionSepArg, NULL},
    {"-ec", ".emptyColor", XrmoptionSepArg, NULL},
    {"-ac", ".activeColor", XrmoptionSepArg, NULL},
    {"-hc", ".hibernateColor", XrmoptionSepArg, NULL},
    {"-swc", ".switchColor", XrmoptionSepArg, NULL},
    {"-pc", ".powerColor", XrmoptionSepArg, NULL},
    {"-sc", ".safeColor", XrmoptionSepArg, NULL},
};

static XtActionsRec actions[] ={
  {"quit",(XtActionProc)quitAction},
  {"toggleslot",(XtActionProc)toggleSlot},
  {"placemenu",(XtActionProc)placeMenu},
};

String wm_trans = "<ClientMessage>WM_PROTOCOLS: quit()\n";
char *program_name;
Atom wm_delete_window;
Atom wm_protocols;

/* ======================================= */

int main(argc, argv)
int argc;
char **argv;
{
    int r, i;
    char ansbuf[BUFSIZ];
    struct sigaction sigact;

    static String fallback_resources[]={
	"*MenuButton.translations:<EnterWindow>: highlight()\\n\
		<LeaveWindow>: reset()\\n\
		Any<BtnDown>: reset() PopupMenu() placemenu()",
	"*baseTranslations: #override\\n\
		<Btn3Down>: quit()\\n\
		<Key>q: quit()\\n\
		<Key>Escape: quit()\\n\
		<Key>Return: toggleslot(0)\\n\
		<Key>0: toggleslot(0)\\n\
		<Key>KP_0: toggleslot(0)\\n\
		<Key>1: toggleslot(1)\\n\
		<Key>KP_1: toggleslot(1)\\n\
		<Key>2: toggleslot(2)\\n\
		<Key>KP_2: toggleslot(2)\\n\
		<Key>3: toggleslot(3)\\n\
		<Key>KP_3: toggleslot(3)",
	"*background: " BG_COLOR "",
#ifdef I18N
	"*international: True",
#else
	"*international: False",
#endif
	"*Label.foreground: black",
	NULL,
    };

#ifdef DEBUG_PRINT_RESOURCE
    /* routine to output fallback resources to stdout */
    char str[BUFSIZ], *p,*new;
    for(i=0; fallback_resources[i] != NULL; i++) {
	p = fallback_resources[i]; new= str;
	while(*p != NULL) {
	    if (*p == 'n' && *(p-1) == '\\') {
		*new++ = 'n'; *new++ = '\\'; *new++ = '\n';
	    } else {
		*new++ = *p;
	    }
	    p++;
	}
	*new = '\0'; puts(str);
    }
    exit(0);
#endif

    XtSetLanguageProc(NULL,NULL,NULL);

    top = XtVaAppInitialize(&app_con,APP_CLASS,options,XtNumber(options),&argc,argv,
			    fallback_resources,NULL,0);
    XtGetApplicationResources(top,(caddr_t)&opts,resources,
                          XtNumber(resources),NULL,0);

    if ((myname = strrchr(*argv, '/')) == NULL) {
	myname = *argv;
    } else {
	myname++;
    }
    argc--; argv++;
    if (argc) usage(stderr, 3);
    if (opts.version) {
	fprintf(stdout, "%s version %s (c) 1999 by Yoshishige Arai\n", myname, VERSION);
	fprintf(stdout, "RCS %s\n", rcsid+1); exit(0);
    }
    if (opts.help) usage(stdout, 0);
    if (openSocket() < 0) {
	perror("openSocket");
	exit(1);
    }

    if ((r = communicate(CMD_NR_SLOTS, ansbuf, BUFSIZ)) > 0) {
	nr_slot = atoi(ansbuf);
    }
    if (nr_slot == 0) {
	closeSocket();
	exit(0);
    }

#if XtSpecificationRelease >= 6
    sigact.sa_handler = signalHandler;
    sigemptyset(&sigact.sa_mask);
    sigaddset(&sigact.sa_mask, SIGHUP);
    sigaddset(&sigact.sa_mask, SIGINT);
    sigaddset(&sigact.sa_mask, SIGTERM);
    sigaddset(&sigact.sa_mask, SIGQUIT);
    sigact.sa_flags = 0;
    sigaction(SIGHUP, &sigact, NULL);
    sigaction(SIGINT, &sigact, NULL);
    sigaction(SIGTERM, &sigact, NULL);
    sigaction(SIGQUIT, &sigact, NULL);
    sigHUPid = XtAppAddSignal(app_con, doSigHUP, NULL);
    sigTERMid = XtAppAddSignal(app_con, doSigTERM, NULL);
#else
    signal(SIGHUP,  signalHandler);
    signal(SIGINT,  signalHandler);
    signal(SIGTERM, signalHandler);
    signal(SIGQUIT, signalHandler);
#endif
    getCardInfo(nr_slot, True);
    if (opts.debug) dumpalias(aliasTop);
    if (!top) {
	fprintf(stderr, "%s: cannot start X session\n", myname);
	closeSocket();
	exit(1);
    }

    disp = XtDisplay(top);
    screen = DefaultScreen(disp);
    root_height = DisplayHeight(disp, screen);
    root_width = DisplayWidth(disp, screen);
    setAppData();
    XtAppAddActions(app_con, actions, XtNumber(actions));
    frame = XtVaCreateManagedWidget("frame",formWidgetClass,top,
				    XtNhorizDistance,0,
				    XtNvertDistance,1,
				    XtNwidth,MY_FRAME_WIDTH, NULL);
    for(i = 0; i < nr_slot; i++) makeSlotFrame(i);
    
    XtAppAddInput(app_con, SOCKET, (caddr_t)XtInputReadMask,
		  (XtInputCallbackProc)statChanged, NULL);
    XtRealizeWidget(top);
    XtOverrideTranslations(top,XtParseTranslationTable(wm_trans));
    wm_protocols = XInternAtom(disp,"WM_PROTOCOLS", False);
    wm_delete_window = XInternAtom(disp,"WM_DELETE_WINDOW",False);
    XSetWMProtocols(disp,XtWindow(top),&wm_delete_window, 1);

    XtAppMainLoop(app_con);
    return 0;
}

static void usage(fp, return_code)
FILE *fp;
int return_code;
{
    static char *option_help[] = {
"    -s		hide drive button strings to sense power of the slot",
"    -n		make smaller size without card names",
"    -always	show card name even if hibernating",
"    -f CONFIG	specify an alternate config file to read",
#ifdef I18N
"    -fs FONTSET	label font set",
"    -bfs FONTSET	button font set",
#else
"    -fn FONT	label font",
"    -bfn FONT	button font",
#endif
"    -ec COLOR	label background color of empty state",
"    -ac COLOR	label background color of active state",
"    -hc COLOR	label background color of hibernate state",
"    -swc COLOR	switch button background color",
"    -pc COLOR	drive button background color on power state",
"    -sc COLOR	drive button background color on safe state",
"    -v		show version",
"    -h		this help",
"",
NULL};
    char **cpp;

    fprintf(fp, "\nusage: %s [-toolkitoption ...] [options]  \n", myname);
    fprintf(fp, " options:\n");
    for (cpp = option_help; *cpp; cpp++) {
        fprintf(fp, "%s\n", *cpp);
    }
    exit(return_code);
}

static int openSocket()
{
    int clen;

    umask(077);
    if ((SOCKET = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0) {
	return(-1);
    }
    bzero((char *)&caddr, sizeof(caddr));
    caddr.sun_family = AF_UNIX;
    sprintf(caddr.sun_path, "%s%06d", CLIENT_NAME, getpid());
    clen = SUN_LEN(&caddr);
    if (bind(SOCKET, (struct sockaddr *)&caddr, clen) < 0) {
	return(-1);
     }
    bzero((char *)&saddr, sizeof(saddr));
    saddr.sun_family = AF_UNIX;
    strcpy(saddr.sun_path, SERVER_NAME);
    slen = SUN_LEN(&saddr);
    return(1);
}

static void closeSocket()
{
    close(SOCKET);
    unlink(caddr.sun_path);
}

static int communicate(cmd, ansbuf, buflen)
char *cmd, *ansbuf;
int buflen;
{
    struct timeval timeout;
    fd_set fds;
    int n, k;

    n = strlen(cmd);
    if (n) {
	if (sendto(SOCKET, cmd, n, 0, (struct sockaddr *)&saddr, slen) != n) {
	    perror("send");
	    exit(1);
	}
    }

    if (buflen == 0) return(0);

    FD_ZERO(&fds);
    FD_SET(SOCKET, &fds);
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;
    n = 0;
    if (select(FD_SETSIZE, &fds, NULL, NULL, &timeout) > 0) {
	if (FD_ISSET(SOCKET, &fds)) {
	    k = 0;
	    if ((n = recvfrom(SOCKET, ansbuf, buflen, 0, NULL, &k)) < 0) {
		perror("recvfrom");
		exit(1);
	    }
	}
    }
    *(ansbuf + n) = 0;
    return(n);
}

static int ismatch(key, str)
char *key, *str;
{
    int r;
    regex_t exp;

    if ((r = regcomp(&exp, key, REG_EXTENDED | REG_NOSUB)) == 0) {
	r = regexec(&exp, str, (size_t)0, NULL, 0);
	regfree(&exp);
    }
    return(r == 0);
}

static char *nameAlias(array, card)
struct CardAlias *array;
struct CardInfo *card;
{
    while(array != NULL) {
	if (ismatch(array->manuf, card->manuf) && ismatch(array->vers, card->vers)) {
	    return(array->alias);
	}
	array = array->next;
    }
    return(NULL);
}

static int parseCardName(buf, card)
char *buf;
struct CardInfo *card;
{
    char *a[MAX_NAME_FEILD];
    char *p, *r, *s;
    int i, slot;

    if ((p = strdup(buf)) == NULL) {
	return(-1);
    }
    s = p + strlen(buf) - 1;
    while(s > p && *s == ' ') *s-- = 0;
    for(s = p, i = 0; i < MAX_NAME_FEILD; i++) {
	while(*s && *s == ' ') s++;
	a[i] = s;
	if (*s == 0) continue;
	while(*s && *s != '~') s++;
	if (i < MAX_NAME_FEILD - 1 && *s == '~') {
	    r = s - 1;
	    *s++ = 0;
	    while(r > a[i] && *r == ' ') *r-- = 0;
	}
    }
    slot = atoi(a[0]);
    card->stat = atoi(a[4]);
    strncpy(card->manuf, *a[1] ? a[1] : S_NA, MAX_INFO_LEN);
    strncpy(card->vers, *a[2] ? a[2] : S_NA, MAX_INFO_LEN);
    strncpy(card->drv, *a[3] ? a[3] : S_NA, MAX_INFO_LEN);

    switch (card->stat) {
    case 0:
	strcpy(card->name, S_EMPTY);
	break;
    case 2:
	strcpy(card->name, S_HIBERNATING);
	break;
    default:
	if ((s = nameAlias(aliasTop, card)) == NULL) {
	    snprintf(card->name, MAX_CARD_NAME, "%s (%s)", card->manuf, card->vers);
	} else {
	    strncpy(card->name, s, MAX_CARD_NAME);
	}
	break;
    }
    XtFree(p);
    return(slot);
}

static int getCardIOinfo(slot, card)
int slot;
struct io_desc *card;
{
    int fd;
    char buf[64];

    card->window = slot;
    card->flags = card->start = card->size = 0;
    sprintf(buf, CARD_DEVICE, slot);
    if (0 > (fd = open(buf, O_RDONLY))) return -1;
    if (ioctl(fd, PIOCGIO, card))
	fprintf(stderr, "%s: ioctl (PIOCGIO) error.\n", myname);
    close(fd);
    return 1;
}

static void getCardInfo(slot_n, noset_flag)
int slot_n;
Bool noset_flag;
{
    int i, r;
    
    char cmd[16], buf[BUFSIZ];

    alias_n = readConfig(opts.config_file);
    for(i = 0; i < slot_n; i++) {
	sprintf(cmd, CMD_GET_INFO, i);
	if ((r = communicate(cmd, buf, sizeof(buf))) > 0) {
	    parseCardName(buf, &cardInfo[i]);
	    getCardIOinfo(i, &cardIO[i]);
	    if (opts.debug) dumpname(i);
	}
	if (!noset_flag && (alias_n > 0)) setCardLabel(i, opts.small);
    }
}

static int shellWord(dst, src)
char **dst, **src;
{
    int quote, n;
    char *d, *s;

    d = *dst; s = *src;
    while (*s && (*s == ' ' || *s == '\t')) s++;
    quote = n = 0;
    while (*s) {
	if (quote == 0 && (*s == ' ' || *s == '\t')) break;
	switch(*s) {
	case '"':
	    if (quote)  quote = 0;
	    else	quote = 2;
	    s++;
	    break;
	case '\\':
	    s++;
	    if (*s == 0) break;
	default:
	    *d++ = *s++;  n++;
	    break;
	}
    }
    *d = 0;   *dst = ++d; *src = s;
    return(n);
}

static int readConfig(file)
char *file;
{
    FILE *fp;
    int n, cnt, line, len;
    char buff[1024];
    char *s, *p, *a[10], *dst;
    struct CardAlias *alp;

    if ((fp = fopen(file, "r")) == NULL) {
	return(-1);
    }

    line = cnt = 0;
    while(fgets(buff, sizeof(buff), fp) != NULL) {
	line++;
	*(buff + strlen(buff) - 1) = 0;
	s = buff;
	while(*s && (*s == ' ' || *s == '\t')) s++;
	if (*s == 0 || *s == '#') continue;

	if ((dst = (char *)XtMalloc(strlen(s)*sizeof(char))) == NULL) {
	    return(-1);
	}
	n = 0;
	p = dst;
	len = 0;
	while (*s) {
	    a[n++] = p;
	    len += shellWord(&p, &s);
	}

	if (n == 4 && strcmp(a[0], "alias") == 0) {
	    alp = (struct CardAlias *)XtMalloc(sizeof(struct CardAlias));
	    alp->buff  = dst;
	    alp->alias = a[1];
	    alp->manuf = a[2];
	    alp->vers  = a[3];
	    if (aliasTop == NULL)
		aliasTop = alp;
	    else {
		aliasPrev->next = alp;
		alp->next = NULL;
	    }
	    aliasPrev = alp;
	    cnt++;
	} else {
	    fprintf(stderr, "readConfig() #%d: illegal format.\n", line);
	    XtFree(dst);
	    continue;
	}
    }
    fclose(fp);
    return(cnt);
}

static void freeAliasArray(array)
struct CardAlias **array;
{
    struct CardAlias *p, *chan;

    if (alias_n > 0) {
	p = *array;
	while(p != NULL) {	
	    chan = p->next;
	    XtFree(p->buff);
	    XtFree(p);
	    p = chan;
	}
	*array = NULL;
    }
}

#if XtSpecificationRelease >= 6
static void doSigHUP(data, id)
XtPointer data;
XtSignalId *id;
{
    sigset_t sigmask, savemask;

    sigemptyset(&sigmask);
    sigaddset(&sigmask, SIGHUP);
    sigprocmask(SIG_BLOCK, &sigmask, &savemask);
    freeAliasArray(&aliasTop);
    getCardInfo(nr_slot, opts.small);
    sigprocmask(SIG_SETMASK, &savemask, NULL);
}

static void doSigTERM(data, id)
XtPointer data;
XtSignalId *id;
{
    closeSocket();
    freeAliasArray(&aliasTop);
    XtAppSetExitFlag(app_con);
    exit(0);
}
#endif

static void signalHandler(signal)
int signal;
{
    switch(signal) {
    case SIGINT :
    case SIGTERM :
    case SIGQUIT :
#if XtSpecificationRelease >= 6
	XtNoticeSignal(sigTERMid);
	break;
    case SIGHUP:
	XtNoticeSignal(sigHUPid);
	break;
#else
	closeSocket();
	freeAliasArray(&aliasTop);
	XtDestroyApplicationContext(app_con);
	exit(0);
    case SIGHUP :
	freeAliasArray(&aliasTop);
	getCardInfo(nr_slot, opts.small);
	XSync(disp, False);
	break;
#endif
    }
}

static void dumpname(n)
int n;
{
    printf("\nslot  = %d\n", n);
    printf("stat  = %d\n", cardInfo[n].stat);
    printf("manuf = %s\n", cardInfo[n].manuf);
    printf("vers  = %s\n", cardInfo[n].vers);
    printf("drv   = %s\n", cardInfo[n].drv);
    printf("name  = %s\n", cardInfo[n].name);
}

static void dumpalias(head)
struct CardAlias *head;
{
    while(head != NULL) {
	printf("\nalias = %s\n", head->alias);
	printf("manuf = %s\n", head->manuf);
	printf("vers  = %s\n", head->vers);
	head = head->next;
    }
}

void setAppData()
{
    XtVaGetApplicationResources(top, (caddr_t)&opts,
			      resources, XtNumber(resources), NULL);
    statColor[0] = opts.empty_color;
    statColor[1] = opts.active_color;
    statColor[2] = opts.hibernate_color;
    PowerColor = opts.power_color;
    SafeColor = opts.safe_color;
    textfont = opts.text_font;
    buttonfont = opts.button_font;
}

static void quitAction(w, e, v, n)
Widget w;
XEvent *e;              
String *v;     
Cardinal *n;
{
    closeSocket();
    freeAliasArray(&aliasTop);
    if (e->type == ClientMessage && e->xclient.type == wm_protocols
       && e->xclient.data.l[0] != wm_delete_window)
        return;
#if XtSpecificationRelease >= 6
    XtAppSetExitFlag(XtWidgetToApplicationContext(w));
#else
    XtDestroyApplicationContext(XtWidgetToApplicationContext(w));
    exit(0);
#endif
}

static void toggleSlot(w, e, v, n)
Widget w;
XEvent *e;
String *v;
Cardinal *n;
{
    int slot = atoi(*v);

    if (slot >= nr_slot) return;
    powerButton(w, (XtPointer)slot, NULL);
    XtVaSetValues(slotFrame[slot].pwrbutton,XtNstate,
		  (cardInfo[slot].stat == 1)? True:False,NULL);
}

static void placeMenu(w, e)
Widget w;
XButtonEvent *e;
{
    int i;
    Dimension w1, x;

    XtVaGetValues(top, XtNwidth,&w1, XtNx,&x, NULL);
    for(i = 0; i < nr_slot; i++)
	XtVaSetValues(slotFrame[i].drvinfo,
		      XtNx,e->x_root + ((root_width > (int)(w1+x+50))? -16:-72),
		      XtNy,e->y_root + ((root_height-100 > e->y_root)? 12:-32),
		      NULL);
}

static void powerButton(w, client, call)
Widget w;
XtPointer client, call;
{
    int slot = (int)client;
    char cmd[8];
    
    switch (cardInfo[slot].stat) {
    case 0:	/* empty */
	return;
    case 1:	/* active */
	sprintf(cmd, CMD_PWR_OFF, slot);
	break;
    case 2:	/* hibernate */
	sprintf(cmd, CMD_PWR_ON, slot);
	XtVaSetValues(slotFrame[slot].drvbutton, XtNbackground,PowerColor, NULL);
	break;
    }
    communicate(cmd, NULL, 0);
}

static void setCardLabel(slot, small)
int slot;
Bool small;
{
    int i, k, n, new_n;
    char buf[BUFSIZ], *p;
    Dimension wt,wo,ht;

    XtVaGetValues(top,XtNheight,&ht,XtNwidth,&wt,NULL);
    XtUnmapWidget(frame);
    XtVaSetValues(slotFrame[slot].drvbutton,
		  XtNlabel,(opts.always_show || cardInfo[slot].stat==1)?
		            cardInfo[slot].drv:S_NA ,
		  XtNbackground,cardInfo[slot].stat == 1 ? PowerColor : SafeColor,
		  NULL);

    if (!small) {
	XtDestroyWidget(slotFrame[slot].label);
	sprintf(buf,"L%d", slot);
	if (opts.always_show || cardInfo[slot].stat==1) {
	    p = cardInfo[slot].name;
	} else if (cardInfo[slot].stat==0) {
	    p = S_EMPTY;
	} else {
	    p = S_HIBERNATING;
	}
	slotFrame[slot].label =
	    XtVaCreateManagedWidget(buf,labelWidgetClass,slotFrame[slot].box,
				    XtNresize,True,
				    XtNfontDEF,textfont,
				    XtNlabel,(String)p,
				    XtNbackground,statColor[cardInfo[slot].stat],
				    XtNborderWidth,LABEL_BORDER_WIDTH,
				    XtNfromHoriz,slotFrame[slot].pwrbutton, NULL);
    }
    XtVaGetValues(slotFrame[slot].box,XtNwidth,&n,NULL);
    XtMapWidget(frame);
    for(i= 0, new_n = 0; i < nr_slot; i++) {
	k = HORIZ_SPACE;
	XtVaGetValues(slotFrame[i].pwrbutton,XtNwidth,&wo,NULL); k+= (int)wo;
	XtVaGetValues(slotFrame[i].drvbutton,XtNwidth,&wo,NULL); k+= (int)wo;
	if (!small) {
	    XtVaGetValues(slotFrame[i].label,XtNwidth,&wo,NULL); k+= (int)wo;
	}
 	if (new_n < k) new_n = k;
    }
    if (16 < (new_n -n)*(new_n -n))
	XtResizeWidget(top, new_n+HORIZ_SPACE, ht);
}

static void statChanged(client, socket, xid)
Widget client;
int *socket;
XtInputId *xid;
{
    int k, n, slot;
    char ansbuf[BUFSIZ];
    struct CardInfo card;

    k = 0;
    n = recvfrom(*socket, ansbuf, BUFSIZ, 0, NULL, &k);
    if (n == 0) return;

    *(ansbuf + n) = '\0';
    slot = parseCardName(ansbuf, &card);
    getCardIOinfo(slot, &cardIO[slot]);
    if (opts.debug) dumpname(slot);
    if (card.stat == 2)
	cardInfo[slot].stat = card.stat;
    else
	cardInfo[slot] = card;

    setCardLabel(slot, opts.small);
    switch(cardInfo[slot].stat) {
    case 1:
	XtSetSensitive(slotFrame[slot].drvbutton, True);
	XtVaSetValues(slotFrame[slot].pwrbutton,XtNstate,False,NULL);
	break;
    case 0:
	XtSetSensitive(slotFrame[slot].drvbutton, False);
    case 2:
	if (opts.drive_state)
	    XtSetSensitive(slotFrame[slot].drvbutton, False);
	break;
    }
}

static void makeSlotFrame(num)
int num;
{
    char name[16], msg[256];
    Boolean s;
#ifdef I18N
    XFontStruct *tmpfs;
    char **ml;
    XFontStruct **fs_list;
#endif

    sprintf(name, "box%d", num);
    slotFrame[num].box =
	XtVaCreateManagedWidget(name,boxWidgetClass,frame,
				XtNorientation,XtorientHorizontal,
				XtNresize,True, XtNborderWidth,0,
				XtNhSpace,1, XtNvSpace,1, XtNwidth,MY_FRAME_WIDTH,
				NULL);
    if (num)
	XtVaSetValues(slotFrame[num].box, XtNfromVert,slotFrame[num-1].box,NULL);
    /* create power button */
    sprintf(msg, " %d ", num);
    sprintf(name, "P%d", num);
    slotFrame[num].pwrbutton = 
	XtVaCreateManagedWidget(name, toggleWidgetClass,slotFrame[num].box,
				XtNlabel,(String)msg,
				XtNborderWidth,1,
 				XtNfontDEF,buttonfont,
				XtNhighlightThickness,2,
				XtNbackground,opts.switch_color,NULL);
    if (cardInfo[num].stat != 1) {
	s = False;
 	XtVaSetValues(slotFrame[num].pwrbutton,XtNstate,&s,NULL);
    }
    /* create drv button */
    sprintf(name, "D%d", num);
    slotFrame[num].drvbutton =
	XtVaCreateManagedWidget(name,menuButtonWidgetClass,slotFrame[num].box,
				XtNresize,True,
 				XtNfontDEF,buttonfont,
				XtNlabel,(String)cardInfo[num].drv,
				XtNborderWidth,1,
				XtNbackground,cardInfo[num].stat == 1 ? PowerColor:SafeColor,
 				XtNhighlightThickness,2,
				XtNfromHoriz,slotFrame[num].pwrbutton,
				NULL);
    sprintf(name, "menu");
    slotFrame[num].drvinfo =
	XtVaCreatePopupShell(name,simpleMenuWidgetClass,slotFrame[num].drvbutton,
 			     XtNbackingStore,NotUseful, XtNsaveUnder,False,
			     XtNwidth,280, XtNheight,20,
			     XtNresize,True, NULL);
    sprintf(msg, "( Slot%d ) flags 0x%03x port 0x%3x size %d bytes ",
	    cardIO[num].window, cardIO[num].flags, cardIO[num].start, cardIO[num].size);
    sprintf(name, "LI%d", num);
    bsb=XtVaCreateManagedWidget(name,smeBSBObjectClass,slotFrame[num].drvinfo,
				XtNfontDEF,textfont,
				XtNlabel,(String)msg,NULL);
#ifdef I18N
    XFontsOfFontSet(textfont,&fs_list,&ml);
    tmpfs = fs_list[0];
    XtVaSetValues(slotFrame[num].drvinfo, XtNwidth,
		  XTextWidth(tmpfs, msg, strlen(msg)) +8,
		  XtNheight, tmpfs->ascent+tmpfs->descent+6, NULL);
#else
    XtVaSetValues(slotFrame[num].drvinfo, XtNwidth,
		  XTextWidth(textfont, msg, strlen(msg)) +8,
		  XtNheight, textfont->ascent+textfont->descent+6, NULL);
#endif
    /* create label */
    if (!opts.small) {
	sprintf(name, "L%d", num);
	slotFrame[num].label =
	    XtVaCreateManagedWidget(name,labelWidgetClass,slotFrame[num].box,
				    XtNresize,True,
				    XtNfontDEF,textfont,
				    XtNlabel,cardInfo[num].name,
				    XtNbackground,statColor[cardInfo[num].stat],
				    XtNborderWidth,LABEL_BORDER_WIDTH,
				    XtNfromHoriz,slotFrame[num].pwrbutton, NULL);
    }
    XtAddCallback(slotFrame[num].pwrbutton,XtNcallback,powerButton,(XtPointer)num);
}
